/* function.cpp -- Copyright (c) 2004 Michael Pyne <michael.pyne@kdemail.net>
 * This software is licensed under the terms of the GNU General Public License,
 * v2 (or any later version).  This software is distributed WITH ABSOLUTELY NO
 * WARRANTY!.
 */
#include <kdebug.h>

#include <qmap.h>
#include <qdict.h>
#include <qstring.h>
#include <qregexp.h>

#include <math.h>
#include "function.h"

// This maps string names to values.
typedef QMap<QString, number_t> valueMap;

// This function pre-defines some constants in the map of values.
valueMap initValueMap()
{
    valueMap map;

    map.insert("pi", M_PI);
    map.insert("e", M_E);

    return map;
}

static valueMap values = initValueMap();

// The parser sets this variable at the end of a successful parse.
number_t parse_result = 0.0;

// Used by the last_parse_status_result() function
int _last_parse_status_result = 0;

static Abakus::TrigMode m_trigMode = Abakus::Degrees;

// Definition of a function that maps one number value onto another number
// of the same type.
typedef number_t (*function_t)(const number_t);

// Hash table of functions.
static QDict<Function> functionDict;

// Dummy return value to enable static initialization
bool addToDict(const char *name, function_t fn)
{
    Function *newFn = new Function;
    QRegExp returnTrigRE("^a(cos|sin|tan)");
    QRegExp needsTrigRE("^(cos|sin|tan)");
    QString fnName(name);

    newFn->name = name;
    newFn->fn = fn;
    newFn->returnsTrig = fnName.contains(returnTrigRE);
    newFn->needsTrig = fnName.contains(needsTrigRE);

    functionDict.insert(name, newFn);

    // Let the dictionary automatically delete the functions.
    functionDict.setAutoDelete(true);

    return false;
}

// Declares a function, with the following parameters:
// * name - The user accessible name of the function
// * fn - The base name of the implementing function.
// * suffix - Suffix to add to the end of the function name
//   Example: f for float functions, l for long double functions, nothing for
//            double functions.
//
// Not only will the function be added to the function hash table, but an
// appropriate function prototype will be generated for broken system that
// don't do it for us.
#define DECLARE_FUNC(name, fn, suffix) bool dummy##name = addToDict(#name, fn##suffix); \
number_t fn##suffix(const number_t);

// Declares a function defined by name, with fn == base name of the
// implementing function.
#define DECLARE_FUNC2(name, fn) DECLARE_FUNC(name, fn, l)

// Declares a function name that is implemented by the function of the
// same base name.
#define DECLARE_FUNC1(name) DECLARE_FUNC2(name, name)

DECLARE_FUNC1(sin);
DECLARE_FUNC1(cos);
DECLARE_FUNC2(abs, fabs);  // math.h version is called fabs
DECLARE_FUNC1(exp);
DECLARE_FUNC1(acos);
DECLARE_FUNC1(asin);
DECLARE_FUNC1(atan);
DECLARE_FUNC1(tan);
DECLARE_FUNC1(cosh);
DECLARE_FUNC1(sinh);
DECLARE_FUNC1(tanh);
DECLARE_FUNC2(ln, log);    // math.h version is called log
DECLARE_FUNC2(log, log10); // math.h version is called log10
DECLARE_FUNC1(sqrt);
DECLARE_FUNC1(ceil);
DECLARE_FUNC1(floor);

const Function *function(const char *name)
{
    return functionDict[name];
}

/* Looks up the named variable (or predefined variable).  Note that although
 * values such as pi are predefined, they can be written over.  This should
 * probably be fixed in the future.  0.0 is returned if the variable doesn't
 * exist, use isValueSet() to see if a value has been set.
 */
number_t value(const char *name)
{
    return values[name];
}

// Returns true if the named value has already been set, false otherwise.
bool isValueSet(const char *name)
{
    return values.contains(name);
}

// Returns true if the named identifier is a function, false otherwise.
bool isFunction(const char *name)
{
    return function(name) != 0;
}

// Sets the named variable name to value, creating the variable if it doesn't
// already exist.
void setValue(const char *name, const number_t value)
{
    values.insert(name, value);
}

// Applies the function identified by func, using value as a parameter.
number_t evaluateFunction(const Function *func, const number_t value)
{
    number_t tempValue(value), tempResult;

    if(func->needsTrig && trigMode() == Abakus::Degrees)
	tempValue *= (M_PI / 180.0);

    tempResult = func->fn(tempValue);

    if(func->returnsTrig)
	tempResult *= (180.0 / M_PI);

    return tempResult;
}

// Returns the status of the last parse operator.  non-zero indicates error,
// 0 indicates success.
int last_parse_status_result(void)
{
    return _last_parse_status_result;
}

void setTrigMode(Abakus::TrigMode mode)
{
    m_trigMode = mode;
}

Abakus::TrigMode trigMode()
{
    return m_trigMode;
}
