#include "mainwindow.h"
#include "abakuscommon.h"

#include <kconfig.h>
#include <kglobal.h>
#include <kconfigbase.h>

#include <qlayout.h>
#include <qradiobutton.h>
#include <qbuttongroup.h>

#include "function.h"
#include "error_handler.h"
#include "resultlistboxtext.h"

MainWindow::MainWindow() : KMainWindow(0), m_popup(0), m_insert(false)
{
    QWidget *box = new QWidget(this);
    QVBoxLayout *layout = new QVBoxLayout(box);
    m_layout = layout;
    layout->setSpacing(6);
    layout->setMargin(6);

    setCentralWidget(box);

    QWidget *configBox = new QWidget(box);
    layout->addWidget(configBox);

    QHBoxLayout *configLayout = new QHBoxLayout(configBox);
    m_showHistory = new QCheckBox(i18n("Show &history"), configBox);
    configLayout->addWidget(m_showHistory);
    connect(m_showHistory, SIGNAL(toggled(bool)), SLOT(slotShowHistory(bool)));

    configLayout->addStretch(1);

    QButtonGroup *buttonGroup = new QButtonGroup(0);

    m_degrees = new QRadioButton(i18n("&Degrees"), configBox);
    configLayout->addWidget(m_degrees);
    buttonGroup->insert(m_degrees);
    slotDegrees();
    connect(m_degrees, SIGNAL(clicked()), SLOT(slotDegrees()));

    m_radians = new QRadioButton(i18n("&Radians"), configBox);
    configLayout->addWidget(m_radians);
    buttonGroup->insert(m_radians);
    connect(m_radians, SIGNAL(clicked()), SLOT(slotRadians()));

    m_history = new QVBox(box);
    layout->addWidget(m_history);
    m_history->setSpacing(6);
    m_history->setMargin(0);

    QLabel *label = new QLabel(i18n("History: "), m_history);
    label->setAlignment(AlignCenter);

    m_result = new KListBox(m_history);
    m_result->installEventFilter(this);
    m_result->setSelectionMode(QListBox::NoSelection);
    m_result->setHScrollBarMode(QListBox::AlwaysOff);

    m_history->setStretchFactor(m_result, 1);
    layout->setStretchFactor(m_history, 1);

    QHBox *editBox = new QHBox(box);
    layout->addWidget(editBox);
    editBox->setSpacing(6);

    m_edit = new KComboBox(editBox);
    m_edit->setFocus();
    m_edit->setEditable(true);
    //m_edit->setInsertionPolicy(KComboBox::AtTop);
    m_edit->setDuplicatesEnabled(false);
    editBox->setStretchFactor(m_edit, 1);

    KPushButton *evalButton = new KPushButton(i18n("&Evaluate"), editBox);

    connect(evalButton, SIGNAL(clicked()), SLOT(slotEvaluate()));

    connect(m_edit, SIGNAL(returnPressed(const QString &)), SLOT(slotReturnPressed(const QString &)));
    connect(m_edit, SIGNAL(textChanged(const QString &)), SLOT(slotTextChanged(const QString &)));
}

bool MainWindow::event(QEvent *e)
{
    ErrorEvent *errorEvent = dynamic_cast<ErrorEvent *>(e);
    if(!errorEvent)
	return KMainWindow::event(e);

    m_lastError = errorEvent->message();
    return true;
}

bool MainWindow::eventFilter(QObject *o, QEvent *e)
{
    if(!o || !e)
	return false;

    if(o == m_result && dynamic_cast<QResizeEvent *>(e))
	m_result->triggerUpdate(true);

    return false;
}

bool MainWindow::queryExit()
{
    saveConfig();
    return KMainWindow::queryExit();
}

void MainWindow::contextMenuEvent(QContextMenuEvent *e)
{
    helpMenu(QString::null, false)->popup(e->globalPos());
}

void MainWindow::polish()
{
    KMainWindow::polish();
    loadConfig();
}

void MainWindow::slotReturnPressed(const QString &text)
{
    QString str = text;

    /* Add right parentheses as needed to balance out the expression. */
    int parenLevel = getParenthesesLevel(str);
    for(int i = 0; i < parenLevel; ++i)
	str += ')';
    
    m_lastError = QString::null;
    number_t result = parseString(str.latin1());

    QString resultVal;
    ResultListBoxText *after = 0;
    QListBoxItem *item;

    if(m_result->count() > 0)
	after = static_cast<ResultListBoxText *>(m_result->item(m_result->count() - 1));

    if(last_parse_status_result() == 0) {
	resultVal = QString::number(result, 'g', 18);

	setValue("ans", result);
	m_insert = true;
    }
    else {
	// We have an ErrorEvent pending, process it first.
	qApp->processEvents();

	resultVal = m_lastError;
	m_insert = false;
    }

    item = new ResultListBoxText(m_result, str, resultVal, after, !m_insert);

    if(m_showHistory->isChecked())
	m_edit->setCurrentText(text);
    else
	m_edit->setEditText(resultVal);

    m_result->setCurrentItem(item);
    m_result->ensureCurrentVisible();

    QTimer::singleShot(0, m_edit->lineEdit(), SLOT(selectAll()));
}

void MainWindow::slotTextChanged(const QString &str)
{
    if(str.length() == 1 && m_insert) {
	QRegExp re("^[-+/*]");
	m_insert = false;
	if(str.find(re) != -1)
	    m_edit->lineEdit()->setText("ans " + str + " ");
    }
}

void MainWindow::slotEvaluate()
{
    slotReturnPressed(m_edit->currentText());
}

void MainWindow::slotShowHistory(bool showHistory)
{
    if(showHistory && m_history->isShown())
	return;

    if(!showHistory && !m_history->isShown())
	return;

    if(showHistory) {
	m_history->show();
	m_newSize = m_oldSize;
    }
    else {
	m_history->hide();
	m_oldSize = size();
	m_newSize = QSize(0, 0);
    }

    updateGeometry();
    QTimer::singleShot(0, this, SLOT(slotUpdateSize()));
}

void MainWindow::slotUpdateSize()
{
    if(m_newSize != QSize(0, 0))
	resize(m_newSize);
    else
	resize(width(), minimumSize().height());
}

void MainWindow::slotDegrees()
{
    setTrigMode(Abakus::Degrees);
}

void MainWindow::slotRadians()
{
    setTrigMode(Abakus::Radians);
}

int MainWindow::getParenthesesLevel(const QString &str)
{
    int level = 0;

    for(unsigned i = 0; i < str.length(); ++i)
	if(str[i] == '(')
	    ++level;
	else if(str[i] == ')')
	    --level;

    return level;
}

void MainWindow::loadConfig()
{
    KConfigGroup config(KGlobal::config(), "Settings");

    bool showHistory = config.readBoolEntry("Show history", true);
    m_showHistory->setChecked(showHistory);
    slotShowHistory(showHistory);

    QString mode = config.readEntry("Trigonometric mode", "Degrees");
    if(mode == "Degrees") {
	setTrigMode(Abakus::Degrees);
	m_degrees->setChecked(true);
    }
    else {
	setTrigMode(Abakus::Radians);
	m_radians->setChecked(true);
    }
}

void MainWindow::saveConfig()
{
    KConfigGroup config(KGlobal::config(), "Settings");

    config.writeEntry("Show history", m_showHistory->isChecked());

    if(trigMode() == Abakus::Degrees)
	config.writeEntry("Trigonometric mode", "Degrees");
    else
	config.writeEntry("Trigonometric mode", "Radians");
}

#include "mainwindow.moc"
