/* parser.y -- Bison definition file for the parser for qt-calc.
 * By Michael Pyne <michael.pyne@kdemail.net>
 * Copyright (c) 2004.
 *
 * You may distribute and modify this file under the terms of the GNU
 * General Public License, v2 (or later).  Notably, this program is distributed
 * WITH ABSOLUTELY NO WARRANTY.
 */

%{

/* Add necessary includes here. */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "function.h"
#include "error_handler.h"

extern char *yytext;

/* Used for strtod checks */
char *endptr;

int yylex(void);
int yyerror(const char *);

%}

%union {
    number_t value;
    Function *fn;
    char *ident;
}

%token <value> NUM
%type <value> EXP FACTOR TERM S EXPONENT NUMBER
%type <fn> FUNC
%token <fn> FN
%token <ident> ID
%type <ident> IDENT
%token POWER "**"

%%

S: EXP { $$ = $1; parse_result = $1; }
S: FUNC '=' {
    post_error_message("You can't assign to function %1.", $1->name);
    YYABORT;
}
S: IDENT '=' EXP {
    setValue($1, $3);
    free($1); /* This variable was strdup'ed, we're done with it. */
    $$ = $3;
    parse_result = $3;
}
S: error {
    post_error_message("Unable to figure out what you're trying to say.%1", "");
    YYABORT;
}

EXP: EXP '+' FACTOR { $$ = $1 + $3; }
EXP: EXP '-' FACTOR { $$ = $1 - $3; }
EXP: FACTOR { $$ = $1; }

FACTOR: FACTOR '*' EXPONENT { $$ = $1 * $3; }
FACTOR: FACTOR '/' EXPONENT { $$ = $1 / $3; }
FACTOR: EXPONENT { $$ = $1; }

EXPONENT: TERM POWER EXPONENT { $$ = powl($1, $3); }
EXPONENT: TERM { $$ = $1; }

TERM: '+' NUMBER { $$ = $2; }
TERM: '+' IDENT { $$ = value($2); free($2); }
TERM: '-' NUMBER { $$ = - $2; }
TERM: '-' IDENT { $$ = - value($2); free($2); }
TERM: '(' EXP ')' { $$ = $2; }

TERM: NUMBER { $$ = $1; }

NUMBER: NUM {
    $$ = strtod(yytext, &endptr);
    if(yytext == endptr) {
	post_error_message ("Can't form a number out of %1", yytext);
	YYABORT;
    }
}

TERM: FUNC TERM {
    $$ = evaluateFunction($1, $2);
}

TERM: IDENT {
    if(isValueSet($1))
	$$ = value($1);
    else {
	post_error_message ("Unknown variable %1", $1);
	free($1);
	YYABORT;
    }

    free($1);
}

IDENT: ID {
    $$ = strdup(yytext);
}

FUNC: FN {
    /* No check necessary, the lexer has already checked for us. */
    $$ = const_cast<Function *>(function(yytext));
}
